/*
 * This software is the confidential and proprietary information of
 * Smardec. You shall not disclose or distribute such Confidential Information.
 * This software shall not be integrated, copied (partially or fully),
 * or reverse engineered into any other software, program, or code that is
 * not intended for Smardec. You shall use it only in accordance with the terms
 * of the license agreement you entered into with Smardec.
 *
 * Copyright (C) 2004-2006 Smardec. All rights reserved.
 *
 * http://www.smardec.com
 *
 */

package com.smardec.demo.j2native.demos;

import com.smardec.asc.hyperlink.HyperlinkUtil;
import com.smardec.demo.j2native.DemoableAdapter;
import com.smardec.demo.j2native.Util;
import com.smardec.demo.j2native.tools.SysInfoUtil;
import com.smardec.demo.j2native.win32.cdplayer.CDPlayer;

import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.io.File;
import java.net.URL;
import java.util.Vector;


public class WinPlayerDemo extends DemoableAdapter {
    private final URL cdpScreenUrl = WinPlayerDemo.class.getClassLoader().getResource("com/smardec/demo/j2native/img/cdplayer/cdp.png");

    public final static String CONSTRAINT_WIN_PLAYER = "WinPlayer";
    public final static String CAPTION_WIN_PLAYER = "Win player";
    public final static String INTRO_WIN_PLAYER = "<html><head>" + css + "</head><body>" +
            "<p align=left color=black>CD player is the program used to play audio discs on your CD-ROM. To launch the program hit the Run button and then you will see the program window on the computer screen and the icon in the system tray.  A double-click on the icon will either hide the player window or show it.</p>" +
            "</body></html>";
    public final static String URL_API_REFERENCE = "player.html";
    public final static String URL_API_REFERENCE2 = "player2.html";
    public final static String API_REFERENCE_PATH = "demo-resources/api-reference/Shell_NotifyIcon.html";
    public final static String API_REFERENCE_PATH2 = "demo-resources/api-reference/NOTIFYICONDATA.html";

    public final static String SRC_WIN_PLAYER = "<html><head>" + css + "</head><body>" +
            "<p align=left color=black>Lets create such a procedure as an example which sends messages to system tray area. The procedure gets the handle of the window associated with the icon, type of the action to be taken, handle to the icon, tip and icon uid." +
            "<br>But first of all, we should create structures which will be used in our procedure. And for this, it is necessary to turn to the description of <i>Shell_NotifyIcon()</i> function." +
            "<p align=right color=black><a href=" + URL_API_REFERENCE + ">View Shell_NotifyIcon() API Reference (in browser)</a></p>" +
            "<br>If you looked attentively, you must have noticed that Shell_NotifyIcon() function uses NOTIFYICONDATA structure, and it means that a wrapper for it should be necessarily created. And again we should refer to MSDN" +
            "<p align=right color=black><a href=" + URL_API_REFERENCE + ">View NOTIFYICONDATA API Reference (in browser)</a></p>" +
            "<br>Now we have all the information we need to create the necessary procedure. And this is what we have as a result:" +
            "<br><pre>    /**\n" +
            "     * Class that is used to perform union functionality. It is used in NOTIFYICONDATA structure\n" +
            "     */\n" +
            "    private static class IconDataUnion extends Union {\n" +
            "        public UInt uTimeout;\n" +
            "        public UInt uVersion;\n" +
            "\n" +
            "        public IconDataUnion() {\n" +
            "            uTimeout = new UInt();\n" +
            "            uVersion = new UInt();\n" +
            "            initialize(new Argument[]{uTimeout, uVersion});\n" +
            "        }\n" +
            "    }\n" +
            "    /**\n" +
            "     * This a wrapper class for system defined structure NOTIFYICONDATA\n" +
            "     */\n" +
            "    private static class NOTIFYICONDATA extends Structure {\n" +
            "        public UInt cbSize;\n" +
            "        public Pointer.Void hWnd;\n" +
            "        public UInt uID;\n" +
            "        public UInt uFlags;\n" +
            "        public UInt uCallbackMessage;\n" +
            "        public Pointer.Void hIcon;\n" +
            "        public AnsiString szTip;\n" +
            "        public UInt dwState;\n" +
            "        public UInt dwStateMask;\n" +
            "        public AnsiString szInfo;\n" +
            "        public IconDataUnion union;\n" +
            "        public AnsiString szInfoTitle;\n" +
            "        public UInt dwInfoFlags;\n" +
            "\n" +
            "        public NOTIFYICONDATA() {\n" +
            "            cbSize = new UInt();\n" +
            "            hWnd = new Pointer.Void();\n" +
            "            uID = new UInt();\n" +
            "            uFlags = new UInt();\n" +
            "            uCallbackMessage = new UInt();\n" +
            "            hIcon = new Pointer.Void();\n" +
            "            szTip = new AnsiString(\"\", 127);\n" +
            "            dwState = new UInt();\n" +
            "            dwStateMask = new UInt();\n" +
            "            szInfo = new AnsiString(\"\", 255);\n" +
            "            union = new IconDataUnion();\n" +
            "            szInfoTitle = new AnsiString(\"\", 63);\n" +
            "            dwInfoFlags = new UInt();\n" +
            "\n" +
            "            initialize(new Argument[]{cbSize, hWnd, uID, uFlags, uCallbackMessage, hIcon, szTip,\n" +
            "                    dwState, dwStateMask, szInfo, union, szInfoTitle, dwInfoFlags});\n" +
            "        }\n" +
            "    }\n" +
            "    /**\n" +
            "     * This function is used to send messages to system tray area.\n" +
            "     *\n" +
            "     * @param hWnd      - Handle to the window that receives notification messages.\n" +
            "     * @param dwMessage - Variable that specifies the action to be taken.\n" +
            "     * @param hIcon     - Handle to the icon to be added, modified, or deleted.\n" +
            "     * @param tip       - Pointer to a null-terminated string with the text for a standard ToolTip\n" +
            "     * @param uidIcon   - Application-defined identifier of the taskbar icon\n" +
            "     */\n" +
            "    private void sendTrayMessage(Pointer.Void hWnd, UInt dwMessage, Pointer.Void hIcon,\n" +
            "                                 String tip, int uidIcon) {\n" +
            "        Library shell32 = new Library(\"shell32\");\n" +
            "        shell_NotifyIcon = shell32.getFunction(\"Shell_NotifyIconA\");\n" +
            "        NOTIFYICONDATA nid = new NOTIFYICONDATA();\n" +
            "\n" +
            "        nid.cbSize.setValue(nid.getLength());\n" +
            "        nid.hWnd.setValue(hWnd.getValue());\n" +
            "        nid.uID.setValue(uidIcon);\n" +
            "        nid.uFlags.setValue(NIF_ICON | NIF_MESSAGE | NIF_TIP);\n" +
            "        nid.szTip.setValue(tip);\n" +
            "        nid.hIcon.setValue(hIcon.getValue());\n" +
            "        nid.uCallbackMessage.setValue(CDPConstants.PM_NOTIFYICON);\n" +
            "        shell_NotifyIcon.invoke(null, dwMessage, new Pointer(nid));\n" +
            "    }\n" +
            "</pre>" +
            "<br>So this is an easy and simple way of how to work with system tray." +
            "</p>" +
            "</body></html>";

    private JPanel jPanelOptions = null;
    private JPanel jPanelDesc = null;
    private JPanel jPanelSrc = null;

    private CDPlayer player;
    private JButton runCDButton; // by default the button must be enabled

    public String getID() {
        return CONSTRAINT_WIN_PLAYER;
    }

    public String getCaption() {
        return CAPTION_WIN_PLAYER;
    }

    public Vector getContentDescriptions() {
        return null;
    }

    public int getState() {
        return STATE_NONE;
    }

    public JPanel getDemoPanel() {
        JPanel jPanelContents = new JPanel(new GridBagLayout()) {
            public void paint(Graphics g) {
                super.paint(g);
                if (!SysInfoUtil.isWin()) {
                    Font font = g.getFont().deriveFont(Font.BOLD, 20);
                    String str = "To view this demo run it under windows";
                    FontMetrics fontmetrics = getFontMetrics(font);
                    g.setFont(font);
                    g.setColor(Color.black);
                    g.drawString(str, (getWidth() - fontmetrics.stringWidth(str)) / 2, (getHeight() - fontmetrics.getHeight()) / 2);
                }
            }
        };
        TitledBorder titledBorder = BorderFactory.createTitledBorder("Demonstration of the common Windows application creation using J2Native");
        titledBorder.setTitleColor(Color.black);
        jPanelContents.setBorder(titledBorder);

        JPanel jPanelTmp = new JPanel(new GridBagLayout());
        ImageIcon cdpImage = new ImageIcon(cdpScreenUrl);
        JLabel cdImage = new JLabel(cdpImage);
        Util.constrain(jPanelTmp, cdImage,
                0, 1, 1, 1, GridBagConstraints.NONE, GridBagConstraints.CENTER, 1.0, 1.0,
                5, 5, 5, 5);

        runCDButton = new JButton(new AbstractAction("     Run     ") {
            public void actionPerformed(ActionEvent e) {
                if (player == null) {
                    player = new CDPlayer();
                    player.initFunctions();
                }
                new Thread(new Runnable() {
                    public void run() {
                        player.winMain();
                        runCDButton.setEnabled(true);
                    }
                }).start();
                runCDButton.setEnabled(false);
            }
        });
        Util.constrain(jPanelTmp, runCDButton,
                0, 2, 1, 1, GridBagConstraints.NONE, GridBagConstraints.EAST, 0.0, 0.0,
                5, 5, 5, 5);

        Util.constrain(jPanelContents, new JScrollPane(jPanelTmp),
                0, 0, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.NORTHWEST, 1, 1,
                0, 5, 0, 5);
        return jPanelContents;
    }

    public JPanel getOptionPanel() {
        if (jPanelOptions != null)
            return jPanelOptions;
        return null;
    }

    public JPanel getDescriptionPanel() {
        if (jPanelDesc != null)
            return jPanelDesc;

        jPanelDesc = new JPanel(new GridBagLayout());
        Util.constrain(jPanelDesc, new JLabel(INTRO_WIN_PLAYER),
                0, 1, 1, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.NORTHWEST, 1.0, 1.0,
                5, 5, 5, 5);
        Util.constrain(jPanelDesc, Box.createGlue(),
                0, 2, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.NORTHWEST, 1.0, 1.0,
                0, 0, 0, 0);
        return jPanelDesc;
    }

    public JPanel getSrcPanel() {
        if (jPanelSrc != null)
            return jPanelSrc;

        jPanelSrc = new JPanel(new GridBagLayout());
        jPanelSrc.setBorder(BorderFactory.createTitledBorder("Windows: CD Player Code Sample"));

        JEditorPane jEditorPane = new JEditorPane("text/html", SRC_WIN_PLAYER);
        jEditorPane.setEditable(false);
        jEditorPane.setCaretPosition(0);
        jEditorPane.setBackground(jPanelSrc.getBackground());
        jEditorPane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent e) {
                if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                    if (URL_API_REFERENCE.equals(e.getDescription())) {
                        try {
                            String s;
                            try {
                                s = new File(API_REFERENCE_PATH).getCanonicalFile().toURI().toString();
                            } catch (NoSuchMethodError er) {
                                s = new File(API_REFERENCE_PATH).getCanonicalFile().toURL().toString();
                                s = convertURL2URI(s);
                            }
                            HyperlinkUtil.openInternetBrowser(s);
                        } catch (Exception e1) {}
                    }
                    if (URL_API_REFERENCE2.equals(e.getDescription())) {
                        try {
                            String s;
                            try {
                                s = new File(API_REFERENCE_PATH2).getCanonicalFile().toURI().toString();
                            } catch (NoSuchMethodError er) {
                                s = new File(API_REFERENCE_PATH2).getCanonicalFile().toURL().toString();
                                s = convertURL2URI(s);
                            }
                            HyperlinkUtil.openInternetBrowser(s);
                        } catch (Exception e1) {}
                    }
                }
            }
        });

        Util.constrain(jPanelSrc, new JScrollPane(jEditorPane),
                0, 1, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.NORTHWEST, 1.0, 1.0,
                5, 5, 5, 5);
        return jPanelSrc;
    }

    public void deinit() {
        if (player != null) {
            player.invokeDestroy();
            try {
                Thread.sleep(500);
            } catch (InterruptedException e1) {
                e1.printStackTrace();
            }
        }
    }

    public void demoWillBecomeVisible() {
        if (runCDButton.isEnabled())
            runCDButton.setEnabled(SysInfoUtil.isWin());
    }
}