/*
 * This software is the confidential and proprietary information of
 * Smardec. You shall not disclose or distribute such Confidential Information.
 * This software shall not be integrated, copied (partially or fully),
 * or reverse engineered into any other software, program, or code that is
 * not intended for Smardec. You shall use it only in accordance with the terms
 * of the license agreement you entered into with Smardec.
 *
 * Copyright (C) 2004-2006 Smardec. All rights reserved.
 *
 * http://www.smardec.com
 *
 */

package com.smardec.demo.j2native.win32.regviewer;

import com.smardec.asc.table.SortedTableHeaderMouseListener;
import com.smardec.asc.table.SortedTableHeaderRenderer;
import com.smardec.asc.table.SortedTableModel;
import com.smardec.demo.j2native.tools.LibraryManager;
import com.smardec.j2native.*;

import javax.swing.*;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.util.ArrayList;
import java.util.Iterator;

public class RegViewer extends JPanel {
    private JTree navTree;
    private DefaultMutableTreeNode root;
    private JTable dataTable;

    private boolean loaded;

    private Function regEnumKeyEx;
    private Function regEnumValue;
    private Function regOpenKey;
    private Function regCloseKey;

    public RegViewer() {
        super(new BorderLayout());
        root = new DefaultMutableTreeNode("My Computer");
        navTree = new JTree(new DefaultTreeModel(root));
        navTree.getSelectionModel().addTreeSelectionListener(new ViewerTreeSelectionListener());
        navTree.addTreeWillExpandListener(new ViewerTreeWillExpandListener());
        dataTable = createDataTable();

        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                new JScrollPane(navTree),
                new JScrollPane(dataTable));

        splitPane.setDividerLocation(250);
        add(splitPane, BorderLayout.CENTER);

        loaded = false;
    }

    public boolean initFunctions() {
        if (loaded) {
            return true;
        }
        try {
            Library advapi32 = new Library("Advapi32");
            LibraryManager.getInstance().registerLibrary(advapi32);
            regEnumKeyEx = advapi32.getFunction("RegEnumKeyExA");
            regEnumValue = advapi32.getFunction("RegEnumValueA");
            regOpenKey = advapi32.getFunction("RegOpenKeyA");
            regCloseKey = advapi32.getFunction("RegCloseKey");

        } catch (LibraryNotFoundException e) {
            e.printStackTrace();
            return false;
        } catch (FunctionNotFoundException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    public void loadRegistry() {
        if (loaded) {
            return;
        }
//    long startTime = System.currentTimeMillis();
        Const hKeyConst;
        RegNode node;
/*
        hKeyConst = new Const(new Pointer.Void(RWConstants.HKEY_CLASSES_ROOT));
        node = new RegNode(new AnsiString("HKEY_CLASSES_ROOT"), hKeyConst);
        root.add(node);
        loadChildNodes(node, hKeyConst, 2);
*/

        hKeyConst = new Const(new Pointer.Void(RWConstants.HKEY_CURRENT_USER));
        node = new RegNode(new AnsiString("HKEY_CURRENT_USER"), hKeyConst);
        root.add(node);
        loadChildNodes(node, hKeyConst, 2);

        hKeyConst = new Const(new Pointer.Void(RWConstants.HKEY_LOCAL_MACHINE));
        node = new RegNode(new AnsiString("HKEY_LOCAL_MACHINE"), hKeyConst);
        root.add(node);
        loadChildNodes(node, hKeyConst, 2);

        hKeyConst = new Const(new Pointer.Void(RWConstants.HKEY_USERS));
        node = new RegNode(new AnsiString("HKEY_USERS"), hKeyConst);
        root.add(node);
        loadChildNodes(node, hKeyConst, 2);

        hKeyConst = new Const(new Pointer.Void(RWConstants.HKEY_CURRENT_CONFIG));
        node = new RegNode(new AnsiString("HKEY_CURRENT_CONFIG"), hKeyConst);
        root.add(node);
        loadChildNodes(node, hKeyConst, 2);
        loaded = true;
        navTree.expandRow(0);
//    System.out.println(" TIME TO LOAD: " + ((System.currentTimeMillis() - startTime) / 1000));
    }

    // -->> for optimized purposes only.... never touch it :) ---
    private UInt dwResult = new UInt();
    private AnsiString name = new AnsiString("", 512);
    private Pointer lpName = new Pointer(name);
    private UInt cbName = new UInt();
    private Pointer lpcbName = new Pointer(cbName);
    private Pointer.Void lpReserved = new Pointer.Void();
    private Pointer.Void lpClass = new Pointer.Void();
    private Pointer.Void lpcbClass = new Pointer.Void();
    private FILETIME ftLastWriteTime = new FILETIME();
    private PointerArgument lpftLastWriteTime = new Pointer.Const(ftLastWriteTime);
    private Pointer.Const pcChildName = new Pointer.Const(new UInt32());
    private Pointer.OutOnly lpNameOutOnly;

    //-----
    private void loadChildNodes(RegNode parentNode, Argument hKey, int levelDown) {
        if (levelDown == 0) {
            return;
        }

//        UInt dwResult = new UInt();
        UInt dwIndex = new UInt(); //
//        AnsiString name = new AnsiString("", 512);
//        Pointer lpName = new Pointer(name);
//        UInt cbName = new UInt();
//        Pointer lpcbName = new Pointer(cbName);
//        Pointer.Void lpReserved = new Pointer.Void();
//        Pointer.Void lpClass = new Pointer.Void();
//        Pointer.Void lpcbClass = new Pointer.Void();
//        FILETIME ftLastWriteTime = new FILETIME();
//        Pointer lpftLastWriteTime = new Pointer(ftLastWriteTime);
        Pointer.Void hkResult = new Pointer.Void();  //
        PointerArgument phkResult = new Pointer(hkResult);  //
        Const dwIndexConst = new Const(dwIndex);    //
        Const hkResultConst = new Const(hkResult);   //
//        Pointer.Const pcChildName = new Pointer.Const(new UInt32());
        while (true) {
            cbName.setValue(512);
            if (lpNameOutOnly == null) {
                regEnumKeyEx.invoke(dwResult, new Argument[]{hKey, dwIndexConst,
                        lpName, lpcbName, lpReserved, lpClass, lpcbClass, lpftLastWriteTime});
                lpNameOutOnly = new Pointer.OutOnly(name);
                lpNameOutOnly.setValue(lpName.getValue());
            } else {
                regEnumKeyEx.invoke(dwResult, new Argument[]{hKey, dwIndexConst,
                        lpNameOutOnly, lpcbName, lpReserved, lpClass, lpcbClass, lpftLastWriteTime});
            }
            if (dwResult.intValue() != RWConstants.ERROR_SUCCESS) {
                break;
            }
            dwIndex.setValue(dwIndex.intValue() + 1);

            AnsiString childName;
            childName = new AnsiString(name.toString().substring(0, Math.min(name.toString().length(), cbName.intValue())));
            pcChildName.setReferencedArgument(childName);
            regOpenKey.invoke(dwResult, hKey, pcChildName, phkResult);
            if (dwResult.intValue() != RWConstants.ERROR_SUCCESS) {
                continue;
            }

            RegNode childNode = new RegNode(childName);
            parentNode.add(childNode);
            loadChildNodes(childNode, hkResultConst, levelDown - 1);
            regCloseKey.invoke(null, hkResultConst);
        }
        if (levelDown > 1) {
            parentNode.setTwoLevelLoaded(true);
        }
    }

    private int closeKey(Argument hKey) {
        UInt uResult = new UInt();
        regCloseKey.invoke(uResult, hKey);
        return uResult.intValue();
    }

    private int openKey(Argument hKey, Pointer.Const childName, PointerArgument hKeyResult) {
        UInt lResult = new UInt();
        regOpenKey.invoke(lResult, hKey, childName, new Pointer(hKeyResult));
        return lResult.intValue();
    }

    private UInt8[] dataMtx;

    // -->> for optimized purposes only.... never touch it :) part 2 ---
//    UInt lResult = new UInt();
//    UInt dwIndex = new UInt();
//    AnsiString valueName = new AnsiString("", 512);
//    Pointer lpValueName = new Pointer(valueName);
//    UInt cValueName = new UInt();
//    Pointer lpcValueName = new Pointer(cValueName);
//    Pointer.Void lpReserved2 = new Pointer.Void();
//    UInt type = new UInt();
//    Pointer lpType = new Pointer(type);
//    Const dwIndexConst = new Const(dwIndex);
//
//    ArrayArgument data;
//
//    Pointer lpData;
//    UInt cData;
//    Pointer lpcbData;
    //-----------------------

    /**
     * This method is used to obtain the values for the specified open registry key
     *
     * @param hKey handle to key to query
     * @return list of RegInfo
     */
    private ArrayList retrieveData(Argument hKey) {
        // prepare arguments
        ArrayList list = new ArrayList();
        UInt lResult = new UInt();
        UInt dwIndex = new UInt();
        AnsiString valueName = new AnsiString("", 512);
        Pointer lpValueName = new Pointer(valueName);
        UInt cValueName = new UInt();
        Pointer lpcValueName = new Pointer(cValueName);
        Pointer.Void lpReserved = new Pointer.Void();
        UInt type = new UInt();
        Pointer lpType = new Pointer(type);
        Const dwIndexConst = new Const(dwIndex);
        if (dataMtx == null) {
            dataMtx = new UInt8[1024];
            for (int i = 0; i < dataMtx.length; i++) {
                dataMtx[i] = new UInt8();
            }
        }
        ArrayArgument data = new ArrayArgument(dataMtx);

        Pointer lpData = new Pointer(data);
        UInt cData = new UInt();
        Pointer lpcbData = new Pointer(cData);

        // create a loop to retrive all values
        while (true) {
            cValueName.setValue(512);
            cData.setValue(1024);
            regEnumValue.invoke(lResult, new Argument[]{hKey, dwIndexConst, lpValueName, lpcValueName,
                    lpReserved, lpType, lpData, lpcbData});
            if (lResult.intValue() != RWConstants.ERROR_SUCCESS) {
                break;
            }
            // increment index
            dwIndex.setValue(dwIndex.intValue() + 1);

            String valueNameStr = valueName.toString().substring(0, cValueName.intValue());
            byte[] bData = new byte[cData.intValue()];
            System.arraycopy(data.toBytesValue(), 0, bData, 0, bData.length);

            // add obtained value to list
            list.add(new RegInfo(valueNameStr, type.intValue(), bData));
        }
        return list;
    }

    private ArrayList getInfoForLastNode(TreePath path) {
        ArrayList info = null;
        if (path.getPathCount() > 1) {
            Argument hKey = ((RegNode) path.getPath()[1]).getHKey();
            if (path.getPathCount() > 2) {
                info = getInfoForLastNode(hKey, path.getPath(), 2);
            } else {
                info = retrieveData(hKey);
            }
        }
        return info;
    }

    private ArrayList getInfoForLastNode(Argument hKey, Object[] path, int index) {
        int errorCode;
        ArrayList info;
        RegNode node = (RegNode) path[index];
        Pointer.Void hkResult = new Pointer.Void();
        Const hkResultConst = new Const(hkResult);
        if ((errorCode = openKey(hKey, new Pointer.Const(node.getName()), hkResult)) != RWConstants.ERROR_SUCCESS) {
            System.err.println("Error #005 - error code: " + errorCode);
            return null;
        }
        index++;
        if (index < path.length) {
            info = getInfoForLastNode(hkResultConst, path, index);
        } else {
            info = retrieveData(hkResultConst);
        }
        if ((errorCode = closeKey(hkResult)) != RWConstants.ERROR_SUCCESS) {
            System.err.println("Error #006 - error code: " + errorCode);
        }
        return info;
    }

    private JTable createDataTable() {
        DefaultTableModel tableModel = new DefaultTableModel(new String[]{"Name", "Type", "Data"}, 0) {
            public boolean isCellEditable(int row, int column) {
                return false;
            }
        };
        SortedTableModel sortedTableModel = new SortedTableModel(tableModel);

        JTable table = new JTable(sortedTableModel) {
            private Color parityColor = new Color(206, 255, 206);
            private Color nonParityColor = new Color(228, 230, 228);

            public TableCellRenderer getCellRenderer(int row, int column) {
                TableCellRenderer cellRenderer = super.getCellRenderer(row, column);
                Component rendererComponent = cellRenderer.getTableCellRendererComponent(dataTable, dataTable.getValueAt(row, column), dataTable.isCellSelected(row, column), false, row, column);
                if (!dataTable.isRowSelected(row)) {
                    rendererComponent.setBackground((row % 2 == 0) ? parityColor : nonParityColor);
                }
                return cellRenderer;
            }
        };
        table.getTableHeader().addMouseListener(new SortedTableHeaderMouseListener(table, sortedTableModel));
        table.getTableHeader().setDefaultRenderer(new SortedTableHeaderRenderer(table, sortedTableModel));

        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.getColumnModel().getColumn(1).setCellRenderer(new TableCellRenderer() {
            JTextArea textArea = new JTextArea();

            public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
                String stringValue = value.toString();
                textArea.setText(stringValue);
                int prefHeight = textArea.getPreferredSize().height;
                if (prefHeight != table.getRowHeight(row)) {
                    table.setRowHeight(row, prefHeight);
                }
                return textArea;
            }
        });
        table.setCellSelectionEnabled(false);
        table.setRowSelectionAllowed(true);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.getColumnModel().getColumn(0).setMinWidth(100);
        table.getColumnModel().getColumn(1).setMinWidth(100);
        table.getColumnModel().getColumn(1).setCellRenderer(new DefaultTableCellRenderer() {
            {
                setHorizontalAlignment(JLabel.CENTER);
            }
        });
        table.getColumnModel().getColumn(2).setMinWidth(120);
        table.getTableHeader().setReorderingAllowed(false);
        return table;
    }

    private class ViewerTreeSelectionListener implements TreeSelectionListener {
        public void valueChanged(TreeSelectionEvent e) {
            if (!loaded) {
                return;
            }
            TreePath path = e.getNewLeadSelectionPath();
            if (path == null) {
                return;
            }
            Object lastPathComp = path.getLastPathComponent();
            DefaultTableModel tableModel = (DefaultTableModel) ((SortedTableModel) dataTable.getModel()).getModel();
            tableModel.setRowCount(0);
            if (lastPathComp instanceof RegNode) {
                StringBuffer stringBuffer = new StringBuffer();
                AnsiString ansiString = new AnsiString();
                Iterator iterator = getInfoForLastNode(path).iterator();
                while (iterator.hasNext()) {
                    RegInfo regInfo = (RegInfo) iterator.next();
                    String type = "";
                    String data = "";
                    switch (regInfo.type) {
                        case RWConstants.REG_NONE:
                            type = new String("REG_NONE");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_SZ:
                            type = new String("REG_SZ");
                            ansiString.fromBytesValue(regInfo.data, 0);
                            data = ansiString.toString();
                            break;
                        case RWConstants.REG_EXPAND_SZ:
                            type = new String("REG_EXPAND_SZ");
                            ansiString.fromBytesValue(regInfo.data, 0);
                            data = ansiString.toString();
                            break;
                        case RWConstants.REG_BINARY:
                            type = new String("REG_BINARY");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_DWORD:
                            type = new String("REG_DWORD");
                            UInt uInt = new UInt();
                            uInt.fromBytesValue(regInfo.data, 0);
                            data = "0x" + Integer.toHexString(uInt.intValue()) + "(" + uInt.intValue() + ")";
                            break;
                        case RWConstants.REG_DWORD_BIG_ENDIAN:
                            type = new String("REG_DWORD_BIG_ENDIAN");
                            Int64 int64 = new Int64();
                            int64.fromBytesValue(regInfo.data, 0);
                            data = "0x" + Long.toHexString(int64.longValue()) + "(" + int64.longValue() + ")";
                            break;
                        case RWConstants.REG_LINK:
                            type = new String("REG_LINK");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_MULTI_SZ:
                            type = new String("REG_MULTI_SZ");
                            int offset = 0;
                            while (offset < regInfo.data.length) {
                                ansiString.fromBytesValue(regInfo.data, offset);
                                stringBuffer.append(ansiString.toString() + "\n");
                                offset += ansiString.getLength();
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_RESOURCE_LIST:
                            type = new String("REG_RESOURCE_LIST");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_FULL_RESOURCE_DESCRIPTOR:
                            type = new String("REG_FULL_RESOURCE_DESCRIPTOR");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_RESOURCE_REQUIREMENTS_LIST:
                            type = new String("REG_RESOURCE_REQUIREMENTS_LIST");
                            for (int i = 0; i < regInfo.data.length; i++) {
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                    }
                    if ("".equals(regInfo.valueName)) {
                        regInfo.valueName = "(Default)";
                    }
                    tableModel.addRow(new Object[]{regInfo.valueName, type, data});
                }
            }
        }
    }

    private class ViewerTreeWillExpandListener implements TreeWillExpandListener {
        public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException {
            if (event.getPath().getPathCount() == 1 || (!loaded)) {
                throw new ExpandVetoException(null);
            }
        }

        public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException {
            if (!loaded) {
                throw new ExpandVetoException(null);
            }
            Object[] path = event.getPath().getPath();
            if ((path.length > 2) && (!((RegNode) path[path.length - 1]).isTwoLevelLoaded())) {
                Window parentFrame = getWindowForComponent(RegViewer.this);
                Cursor oldCursor = null;
                if (parentFrame != null) {
                    oldCursor = parentFrame.getCursor();
                    parentFrame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
                }
                loadSpecificChilds(((RegNode) path[1]).getHKey(), path, 2, 2);
                if (parentFrame != null) {
                    parentFrame.setCursor(oldCursor);
                }
            }
        }

        private void loadSpecificChilds(Argument hKey, Object[] path, int index, int levelDown) {
            int errorCode;
            RegNode node = (RegNode) path[index];
            Pointer.Void hkResult = new Pointer.Void();
            Const hkResultConst = new Const(hkResult);
            if ((errorCode = openKey(hKey, new Pointer.Const(node.getName()), hkResult)) != RWConstants.ERROR_SUCCESS) {
                System.err.println("Error #001 - error code: " + errorCode);
                return;
            }
            index++;
            if (index < path.length) {
                loadSpecificChilds(hkResultConst, path, index, levelDown);
            } else {
                Pointer.Void childHKey = new Pointer.Void();
                Const childHKeyConst = new Const(childHKey);
                Pointer.Const pChildNodeName = new Pointer.Const(new UInt32());
                for (int i = node.getChildCount() - 1; i >= 0; i--) {
                    RegNode childNode = (RegNode) node.getChildAt(i);
                    pChildNodeName.setReferencedArgument(childNode.getName());
                    if ((errorCode = openKey(hkResultConst, pChildNodeName, childHKey)) != RWConstants.ERROR_SUCCESS) {
                        System.err.println("Error #003 - error code: " + errorCode);
                        continue;
                    }

                    loadChildNodes(childNode, childHKeyConst, levelDown);

                    if ((errorCode = closeKey(childHKeyConst)) != RWConstants.ERROR_SUCCESS) {
                        System.err.println("Error #004 - error code: " + errorCode);
                    }
                }
                node.setTwoLevelLoaded(true);
            }
            if ((errorCode = closeKey(hkResultConst)) != RWConstants.ERROR_SUCCESS) {
                System.err.println("Error #002 - error code: " + errorCode);
            }
        }
    }

    private static Window getWindowForComponent(Component parentComponent) {
        if (parentComponent == null) {
            return JOptionPane.getRootFrame();
        }
        if (parentComponent instanceof Frame || parentComponent instanceof Dialog) {
            return (Window) parentComponent;
        }
        return getWindowForComponent(parentComponent.getParent());
    }

    /**
     * Class that holds registry info
     */
    private class RegInfo {
        public String valueName;
        public int type;
        public byte[] data;

        public RegInfo(String valueName, int type, byte[] data) {
            this.valueName = valueName;
            this.type = type;
            this.data = data;
        }
    }

    private class FILETIME extends Structure {
        public UInt dwLowDateTime;
        public UInt dwHighDateTime;

        public FILETIME() {
            dwLowDateTime = new UInt();
            dwHighDateTime = new UInt();

            initialize(new Argument[]{dwLowDateTime, dwHighDateTime});
        }
    }
}
